define(['context'], function (Context) {

  // alias context variable
  var c = Context;

  describe("Context", function () {

    it("should have default values", function () {
      expect(c.driver).toBe(null);
      expect(c.loaded).toBe(false);
      expect(c.meta).toEqual({});
      expect(c.queryMode).toBe("filter");
    });

    it("should return proper quote char", function () {
      c.driver = "mysql";
      expect(c.quoteChar()).toBe('`');

      c.driver = "postgresql";
      expect(c.quoteChar()).toBe('"');

      c.driver = "redshift";
      expect(c.quoteChar()).toBe('"');

      c.driver = null;
      expect(c.quoteChar()).toBe('`');
    });

    describe("queryMatchesCurrentTableSchema", function () {

      it("return true if columns from query matches current table schema", function () {
        c.result = {meta:[{name:"id"}]};
        c.meta.columns = ["id"];
        expect(c.queryMatchesCurrentTableSchema()).toBe(true);

        c.result = {meta:[{name:"id"}]};
        c.meta.columns = ["id", "extra_col"];
        expect(c.queryMatchesCurrentTableSchema()).toBe(true);
      });

      it("return false if any of the columns from query are not present in table schema", function () {
        c.result = {meta:[{name:"id"}, {name:"foo_bar"}]};
        c.meta.columns = ["id"];
        expect(c.queryMatchesCurrentTableSchema()).toBe(false);

        c.result = {meta:[{name:"id1"}]};
        c.meta.columns = ["id2"];
        expect(c.queryMatchesCurrentTableSchema()).toBe(false);
      });

      it("return false if table schema is not present", function () {
        c.result = {meta:[{name:"id"}]};
        c.meta = undefined;
        expect(c.queryMatchesCurrentTableSchema()).toBe(false);

        c.meta = null;
        expect(c.queryMatchesCurrentTableSchema()).toBe(false);

        c.meta = {meta: null};
        expect(c.queryMatchesCurrentTableSchema()).toBe(false);

        c.meta = {meta: []};
        expect(c.queryMatchesCurrentTableSchema()).toBe(false);
      });

      it("return false if query meta is not present", function () {
        c.result = undefined;
        c.meta = undefined;
        expect(c.queryMatchesCurrentTableSchema()).toBe(false);

        c.result = undefined;
        c.meta = {columns:["id"]};
        expect(c.queryMatchesCurrentTableSchema()).toBe(false);
      });

    });

    describe("getTableFromQuery", function () {

      it("should return current table when present and check integrity", function () {

        // table defined, but result data missing
        c.table = "test_table1";
        c.meta = {columns:["id"]};
        c.result = {meta: [{name: "id", "table": "test_table1"}]};
        expect(c.getTableFromQuery()).toBe("test_table1");

        // table defined, but result data has different table name
        c.table = "test_table2";
        c.meta = {columns:["id"]};
        c.result = {meta: [{name: "id", "table": ""}]};
        expect(c.getTableFromQuery()).toBe("test_table2");
      });

      it("should return table from result meta if integrity check fails and result meta is present", function () {

        // table is not defined, but table name can be found in meta
        c.table = undefined;
        c.meta = {};
        c.result = {
          meta: [
            {
              name: "id",
              size: 11,
              table: "foo_bar",
              typeName: "int4"
            }
          ]
        };
        expect(c.getTableFromQuery()).toBe("foo_bar");

        // table is defined, but integrity check fails
        c.table = "random_table";
        c.meta = {};
        c.result = {
          meta: [
            {
              name: "id",
              size: 11,
              table: "foo_bar",
              typeName: "int4"
            }
          ]
        };
        expect(c.getTableFromQuery()).toBe("foo_bar");
      });

      it("should return null when table can't be found", function () {

        // missing table and result data
        c.table = undefined;
        c.result = undefined;
        c.meta = {};
        expect(c.getTableFromQuery()).toBe(null);

        // empty table name in result meta
        c.result = {
          meta: [
            {
              name: "id",
              size: 11,
              typeName: "int4",
              table: ""
            }
          ]
        };
        expect(c.getTableFromQuery()).toBe(null);

        // query columns doesnt match schema
        c.table = "foobar1";
        c.meta = {columns:["id"]};
        c.result.meta = null;
        c.result = {
          meta: [
            {
              name: "blah",
              size: 11,
              typeName: "int4"
            }
          ]
        };
        expect(c.getTableFromQuery()).toBe(null);

        // query columns doesnt match schema, integrity check fails, but result meta has the same table that
        // just failed integrity test
        c.table = "foobar2";
        c.meta = {columns:["id"]};
        c.result.meta = null;
        c.result = {
          meta: [
            {
              name: "blah",
              size: 11,
              typeName: "int4",
              table: "foobar2"
            }
          ]
        };
        expect(c.getTableFromQuery()).toBe(null);

        // missing table from result meta, current table not set
        c.table = null;
        c.meta = null;
        c.result = {
          meta: [
            {
              name: "custom",
              size: 11,
              typeName: "int4"
            }
          ]
        };
        expect(c.getTableFromQuery()).toBe(null);

        // missing meta from result data
        //c.result = {
        //  meta: null
        //};
        //expect(c.getTableFromQuery()).toBe(null);
      });
    });
  });
});