define(['jquery', 'underscore', 'backbone', 'tpls', 'grid', 'events', 'context', 'squel', 'models/query',
      'models/query/filter/condition', 'handlebars', 'editors/datetime', 'inview'],
    function ($, _, Backbone, Templates, Grid, Events, Context, squel, Query, Condition, Handlebars, DateTimeEditor) {
      return Backbone.View.extend({
        template: Templates["content/output/table"],
        initialize: function () {
          this.listenTo(Events, 'query:fetched', this.renderGrid, this);
          this.listenTo(Events, 'table:change', this.clearSorting, this);
          _.bindAll(this, "saveCell", "saveRow", "removeRow", "sortData", "clearSorting", "toggleToolbar",
              "removeRows", "success", "error");
        },
        /**
         * Initial data
         */
        initialData: null,
        textareaEditorTypes: ["MEDIUMTEXT", "LONGTEXT", "VARCHAR", "TEXT", "CHAR"],
        sorting: null,
        events: {
          "click .sensei-grid-toolbar .remove": "removeRows"
        },
        tips: [
          "press {{meta}}+a to select/deselect all rows",
          "press {{meta}}+backspace to delete selected or active rows",
          "press {{meta}}+d to duplicate active row",
          "press enter on active cell to activate editor",
          "select multiple rows by holding down shift while moving with arrows",
          "use arrows and tab/shift+tab to navigate between cells",
          "press escape to close editor/lose focus from grid",
          "press space to select row",
          "press enter/tab with open editor to persist changes",
          "use the empty row at bottom to insert new data",
          "press escape to deselect rows"
        ],
        promises: [],
        /**
         * Sensei Grid instance
         */
        grid: null,
        getTable: function () {

          if (Context.query_mode === "raw") {
            return Context.getTableFromQuery();
          }

          return Context.table;
        },
        /**
         * Choose editor depending on column type
         * @param meta
         * @returns {Array}
         */
        mapColumnEditors: function (meta) {
          return _.map(meta, function (col) {

            var cType = col.typeName.toUpperCase();
            var cSize = parseInt(col.size);
            var meta = Context.getColumnMeta(col.name);
            var rawType = meta && meta.Type ? meta.Type : "";
            var isNull = meta && meta.Null && meta.Null.toLowerCase() === "yes";
            var setValues = rawType.extractSetValues();

            console.log("map column editors:", col.name, cSize, cType, meta, setValues);

            // if field could contain large text, use TextareaEditor
            if (_.contains(this.textareaEditorTypes, cType) && cSize > 255) {
              return {name: col.name, type: "string", editor: "TextareaEditor"};
            }

            // check if column is boolean-ish
            if (cType === "TINYINT" && cSize === 1) {
              return {name: col.name, type: "string", editor: "BooleanEditor"};
            }

            // check if column is suited for datepicker
            if (cType === "DATETIME" || cType === "DATE" || cType === "TIMESTAMP") {
              return {name: col.name, type: "string", editor: "DateTimeEditor"};
            }

            if (setValues && !_.isEmpty(setValues)) {
              if (isNull) {
                setValues.unshift("NULL");
              }
              return {name: col.name, type: "string", editor: "SelectEditor", editorProps: {values: setValues}}
            }

            return {name: col.name, type: "string", editor: "BasicEditor"};
          }, this);
        },
        /**
         * Render grid
         * @param data Object Query data and meta
         */
        renderGrid: function (data) {

          // determine if we are in filter mode, specific features are only available for filter queries
          var isCRUDMode = Context.query_mode === "filter" || !_.isEmpty(Context.getTableFromQuery());

          // load meta from context if it not present in returned result set
          if (_.isEmpty(data.meta)) {
            data.meta = Context.meta["columnSchema"];
          }

          // map columns
          var columns = this.mapColumnEditors(data.meta);

          // deep copy and store initial data
          this.initialData = JSON.parse(JSON.stringify(data.rows));

          // default skipOnDuplicate to null
          var skipOnDuplicate = null;

          if (Context.meta.primary_keys) {
            // skip primary keys on row duplication
            skipOnDuplicate = Context.meta.primary_keys;
          }

          console.log("TableView.renderGrid", "isCRUDMode:", isCRUDMode, "columns:", columns);

          // sensei grid options
          var that = this;
          var options = {
            tableClass: "table table-bordered table-condensed",
            emptyRow: isCRUDMode,
            moveOnRowRemove: true,
            sortable: true,
            removable: isCRUDMode,
            initialSort: this.sorting,
            skipOnDuplicate: skipOnDuplicate,
            selectable: isCRUDMode,
            toolbar: isCRUDMode,
            readonly: !isCRUDMode,
            dateFormat: "yyyy-mm-dd 00:00:00",
            getContainer: function () {
              return that.$(".sensei-grid-table-wrapper");
            }
          };

          // grid initialization
          this.grid = this.$(".sensei-grid").grid(data.rows, columns, options);

          // register basic editor
          if (isCRUDMode) {
            this.grid.registerEditor(BasicEditor);
            this.grid.registerEditor(TextareaEditor);
            this.grid.registerEditor(BooleanEditor);
            this.grid.registerEditor(SelectEditor);
            this.grid.registerEditor(DateTimeEditor);
          }

          // listen to events
          this.grid.events.on("editor:save", this.saveCell);
          this.grid.events.on("row:save", this.saveRow);
          this.grid.events.on("row:remove", this.removeRow);
          this.grid.events.on("cell:select", this.toggleToolbar);
          this.grid.events.on("cell:deactivate", this.toggleToolbar);
          this.grid.events.on("row:mark", this.toggleToolbar);
          this.grid.events.on("row:unmark", this.toggleToolbar);
          this.grid.events.on("column:sort", this.sortData);

          // render grid
          this.grid.render();

          // render toolbar
          this.renderToolbar(data);
        },
        renderToolbar: function (data) {
          // render toolbar and set returned row count
          var count = data && data.rows ? data.rows.length : 0;
          var tpl = Templates["content/output/table_toolbar"]({count: count});
          this.$(".sensei-grid-toolbar").html(tpl);

          // render random tip
          this.renderTip();
        },
        renderTip: function () {
          var tip = _.shuffle(this.tips)[0];
          var tipTpl = Handlebars.compile(tip);
          var ctx = {meta: Context.getMetaKey()};
          this.setTip("Tip: " + tipTpl(ctx));
        },
        toggleToolbar: function () {
          var $active = this.grid.getActiveCell();
          var selectedRows = this.grid.getSelectedRows();

          // update selected row count
          this.$(".selected-row-count").text(selectedRows.length || ($active && $active.length));

          // toggle remove button
          if (($active && $active.length > 0 && !$active.parent().is(".sensei-grid-empty-row")) ||
              selectedRows.length > 0) {
            this.$(".sensei-grid-toolbar .remove").removeClass("disabled");
          } else {
            this.$(".sensei-grid-toolbar .remove").addClass("disabled");
          }

          // toggle duplicate button
          if (($active && $active.length > 0 && !$active.parent().is(".sensei-grid-empty-row")) ||
              selectedRows.length === 1) {
            this.$(".sensei-grid-toolbar .duplicate").removeClass("disabled");
          } else {
            this.$(".sensei-grid-toolbar .duplicate").addClass("disabled");
          }
        },
        /** @param {!jQuery.Event} e */
        removeRows: function (e) {
          e.preventDefault();

          this.setTip("Removing rows...");

          var $active = this.grid.getActiveCell();
          var selectedRows = this.grid.getSelectedRows();

          // reset promises
          this.promises = [];

          if (selectedRows && selectedRows.length > 0) {
            // remove selected rows if any
            var grid = this.grid;

            selectedRows.each(function () {
              grid.removeRow($(this), {avoidReload: true});
            });

            // call this.success when all promises are done
            $.when.apply($, this.promises).done(this.success);

          } else {
            // if rows are not selected, remove active row
            //if (this.grid.$lastActiveCell && this.grid.$lastActiveCell.length === 1) {
            if ($active && $active.length === 1) {
              this.grid.removeRow($active.parent());
              //this.grid.removeRow(this.grid.$lastActiveCell.parent());
            }
          }
        },
        sortData: function (col, order) {
          // trigger data:sort event
          // query builder will respond to this event
          this.sorting = {col: col, order: order};
          Events.trigger("data:sort", col, order);
          this.setTip("Sorting data...");
        },
        clearSorting: function () {
          this.sorting = null;
        },
        saveCell: function (data, $cell) {
          var row = $cell.parent("tr").index();
          if (this.grid.getCellStatus($cell)) {
            this.updateRow(data, row);
          }
        },
        saveRow: function (data, $row) {
          var grid = this.grid;
          this.insertRow(data, function () {
            grid.setRowSaved($row);
          });
        },
        /**
         * Replace null values with string "NULL"
         * @param value String
         * @returns {*}
         */
        parseValue: function (value, skipNull) {
          if (value === null) {
            return "NULL";
          }
          if (!skipNull && !_.isNumber(value) && _.isEmpty(value) && !_.isBoolean(value)) {
            return "NULL";
          }
          var normalizedValue = value.toString().toLowerCase();
          if (_.isBoolean(value) || $.inArray(normalizedValue, ["true", "false"]) > -1) {
            // convert boolean to integer
            return value === true || normalizedValue === "true" ? 1 : 0;
          }
          return value;
        },
        /**
         * Insert row in DB
         * @param values Object
         */
        insertRow: function (values, successCallback) {

          this.setTip("Inserting row...");

          // squel insert instance
          var insert = squel.insert(Context.getSquelSettings()).into(this.getTable());

          // set values
          this.setValues(insert, values, false);

          // execute query
          this.executeQuery(insert, successCallback);
        },
        /**
         * Choose update strategy depending on schema/driver
         * and return key-value pairs for update condition
         * @param row Int
         * @returns {object}
         */
        updateStrategy: function (row) {

          // check if row exists in initialData object
          if (!this.initialData[row]) {
            console.warn("row not found in initial data, can't choose update strategy");
            return false;
          }

          // check if primary keys exist for this table
          if (!_.isEmpty(Context.meta.primary_keys)) {

            var conditions = {};

            // iterate primary keys and set key-value pairs from initialData object
            _.each(Context.meta.primary_keys, function (key) {
              if (_.has(this.initialData[row], key)) {
                conditions[key] = this.initialData[row][key];
              }
            }, this);

            // check if all primary keys are used and was found in initialData object
            if (_.size(conditions) === Context.meta.primary_keys.length) {
              console.log("primary key update strategy can be done", conditions);
              return conditions;
            } else {
              console.warn("primary keys found, but not enough data in initial storage, falling back to " +
                  "whole row update strategy", conditions, Context.meta.primary_keys, row);
            }
          }

          console.log("update strategy chosen: initialData");
          return this.initialData[row];
        },
        /**
         * Update row in DB
         * @param data Object
         * @param row Int
         */
        updateRow: function (data, row) {

          this.setTip("Updating row...");

          // squel update instance
          var update = squel.update(Context.getSquelSettings()).table(this.getTable());

          // set values
          this.setValues(update, data, true);

          // get update strategy
          var updateValues = this.updateStrategy(row);

          // set where condition from update strategy
          _.each(updateValues, function (value, key) {
            // parse value
            value = this.parseValue(value, true);

            var condition = new Condition({column: key, operator: "=", value: value});
            update.where(condition.queryString(), value);
          }, this);

          if (Context.driver === "mysql") {
            // limit update to single row, if db backend supports it
            update.limit(1);
          }

          // execute query
          this.executeQuery(update, function () {
          }, {avoidReload: true});
        },
        removeRow: function (data, row, $row, options) {

          this.setTip("Removing row...");

          console.log("TableView.removeRow", data, row, $row, options);

          // squel delete instance
          var del = squel.delete(Context.getSquelSettings()).from(this.getTable());

          // get update strategy
          var updateValues = this.updateStrategy(row);

          if (!updateValues || _.isEmpty(updateValues)) {
            console.warn("update strategy not found, can't remove row");
            return;
          }

          // set where condition from update strategy
          _.each(updateValues, function (value, key) {
            // parse value
            value = this.parseValue(value, true);

            var condition = new Condition({column: key, operator: "=", value: value});
            del.where(condition.queryString(), value);
          }, this);

          if (Context.driver === "mysql") {
            // limit delete to single row, if db backend supports it
            del.limit(1);
          }

          // execute query
          this.initialData.splice(row, 1);
          this.executeQuery(del, function () {
            console.info("delete success callback: row deleted", del.toString(), options);
          }, options);
        },
        getColumnTypeByName: function (column) {
          var schema = _.find(Context.meta.columnSchema, function (s) {
            return s.name === column;
          });
          return schema["typeName"];
        },
        /**
         * Set field values on squel instance
         * @param squel Squel instance
         * @param values Object
         * @param skipNull
         * @returns squel
         */
        setValues: function (squel, values, skipNull) {

          // iterate through values
          _.each(values, function (value, column) {
            // set value only if the column exists in table or if this is raw query mode
            if (_.contains(Context.meta.columns, column) || Context.query_mode === "raw") {
              squel.set(column, this.parseValue(value, skipNull));
            }
          }, this);
        },
        /**
         * Execute insert/update query
         * @param successCallback
         * @param query Squel query
         */
        executeQuery: function (query, successCallback, options) {
          console.log("executeQuery", query);

          // get parameterized query
          var queryWithParams = query.toParam();

          // prepare post data
          var data = {
            q: queryWithParams.text, "values[]": queryWithParams.values,
            table: this.getTable()
          };

          // query model is used for insert/update queries
          var queryModel = new Query();

          // a special case when we really need this because of successCallback passing issues
          var that = this;

          // persist last active cell
          //var $cell = this.grid.getActiveCell();
          ////var $row = this.grid.getCellRow($cell);
          //var $row = $cell.parent();
          //console.log("setting incorrect pos:", $cell.index(), $row.index());
          //var lastPosition = {cell: $cell.index(), row: $row.index()};

          // modify queryModel settings and change endpoint url
          var prom = queryModel.fetch({
            data: data,
            // insertURL is used for both inserts and updates
            url: queryModel.insertURL(),
            type: "POST",
            success: function (model, res, xhr) {
              if (successCallback) {
                successCallback();
              }

              that.setTip("Done!");
              // don't call success method if skipReload option is passed
              if (!options || !options.avoidReload) {
                that.success(model, res, xhr);
              }
            },
            error: this.error
          });

          this.promises.push(prom);
        },
        success: function () {

          // its safer to reload grid after each action, to ensure
          // initialData and rows are in sync
          // var that = this;
          // Events.trigger("query:execute", function () {
          //     var $row = that.grid.getRowByIndex(lastPosition.row);
          //     var $cell = that.grid.getCellFromRowByIndex($row, lastPosition.cell);
          //     that.grid.setActiveCell($cell);
          //     console.info("SUCCESS CALLBACK", lastPosition, $row, $cell);
          // });
          //var that = this;
          var grid = this.grid;
          var $active = grid.getActiveCell();

          Events.trigger("query:execute", function () {
            var rowIndex = $active.parent().index();
            var cellIndex = $active.index();
            //console.log("after delete, last pos", cellIndex, rowIndex);
            var $row = grid.getRowByIndex(rowIndex);
            var $cell = grid.getCellFromRowByIndex($row, cellIndex);

            if ($cell && $cell.length === 1) {
              grid.setActiveCell($cell);
            }

            //console.log("activate last cell", $cell, $cell.has(":focus"));

            //p, that.grid.$lastActiveCell.parent().index(), that.grid.$lastActiveCell.index()
            //var $row = that.grid.getRowByIndex();
            //var $cell = that.grid.getCellFromRowByIndex($row, lastPosition.cell);
            //that.grid.setActiveCell($cell);
            //console.info("SUCCESS CALLBACK", lastPosition, $row, $cell);
          });
        },
        error: function (model, error) {
          if (error.responseJSON && error.responseJSON.error) {
            this.setTip("Error: " + error.responseJSON.error);
          } else {
            this.setTip("Error: unable to complete action.");
          }
          console.log("queryExecute:error", error["responseJSON"]);
          alert("Error: %(error)s".format(error["responseJSON"]));
        },
        setTip: function (text) {
          this.$(".tip-label").text(text).attr("title", text);
        },
        /**
         * Render view
         * @returns {*}
         */
        render: function () {
          this.$el.html(this.template());

          if (!_.isEmpty(Context.result)) {
            this.renderGrid(Context.result);
          }
          return this;
        }
      });
    });
