define(['jquery', 'underscore', 'backbone', 'tpls', 'models/output/viz/settings',
      'events', 'models/output/viz/chart', 'context', 'collections/output/charts',
      'views/content/output/viz/chart_list_item', 'baseview'],
    function ($, _, Backbone, Templates, SettingsModel, Events, Chart, Context, Charts,
              ChartListItem, BaseView) {
      var SettingsView = BaseView.extend({
        template: Templates["content/output/viz/settings"],
        events: {
          "click .chart-modes .btn": "switchChartType",
          "change .chart-show-legend": "toggleLegend",
          "click .save-chart": "saveChart",
          "show.bs.dropdown .load-chart": "loadChartMenu",
          "change select[name=xaxis]": "updateXAxis",
          "click .embed-url, click .embed-url-wrapper": function (e) {
            $(e.currentTarget).select();
          }
        },
        $chartList: null,
        chartModes: ["Line", "Spline", "Area", "Area Spline", "Bar", "Pie", "Donut", "Scatter", "Step"],
        switchChartType: function (e) {
          // set active item class and remove it from other elements
          $(e.target).addClass("btn-primary").siblings().removeClass("btn-primary");

          // if chart is loaded switch chart type
          if (this.parent.subviews && this.parent.subviews.viz && this.parent.subviews.viz.chart) {
            this.chartMode = this.parseModeTitle($(e.target).data("mode"));
            this.parent.subviews.viz.chart.transform(this.chartMode);
          } else {
            console.warn("can't switch chart type, because chart is not loaded");
          }
        },
        updateXAxis: function (e) {
          Events.trigger("settings:update");
        },
        getXAxis: function () {
          return this.$("select[name=xaxis]").val();
        },
        parseModeTitle: function (title) {
          return title.toLowerCase().replace(" ", "-");
        },
        saveChart: function () {

          var $save_btn = this.$(".save-chart");
          $save_btn.text("Saving...").prop("disabled", true).removeClass("btn-danger");

          // set chart model
          var chart = this.model ? this.model : new Chart();

          // chart settings
          var settings = {
            chartMode: this.getChartMode(),
            showLegend: this.showLegend,
            xaxis: this.getXAxis()
          };

          // model data
          var data = {};
          data.title = this.$(".chartTitle").val().trim() || "Untitled";
          data.conn = Context.connection;
          data.db = Context.db;
          data.query = Context.query;
          data.settings = JSON.stringify(settings);
          data.team_shared = this.$(":input[name=team_shared]").is(":checked");
          data.public = this.$(":input[name=public]").is(":checked");

          var that = this;

          // save chart
          chart.save(data, {
            success: function (model, data) {
              console.log("success: chart saved", data);
              Events.trigger('chart:create', model);
              $save_btn.text("Save").prop("disabled", false);

              // so updating works right away
              that.model = model;

              // toggle embed url
              that.toggleEmbedURL(model);
            },
            error: function (model, xhr) {
              $save_btn.text("Failed! Try again?").prop("disabled", false).addClass("btn-danger");
              if (xhr.responseJSON && xhr.responseJSON.errors) {
                console.warn("error: chart not saved", xhr.responseJSON.errors);
              } else {
                console.warn("unknown error");
              }
            }
          });
        },
        toggleLegend: function () {
          this.showLegend = !this.showLegend;
          Events.trigger("settings:update");
        },
        getShowLegend: function () {
          return this.showLegend;
        },
        getChartMode: function () {
          if (this.chartMode) {
            return this.chartMode.toLowerCase();
          }
          return "line";
        },
        initialize: function () {
          _.bindAll(this, "addChartListItem", "toggleEmbedURL");

          this.charts = new Charts();
          this.listenTo(this.charts, "add", this.addChartListItem);
          this.listenTo(Events, "chart:load", this.loadChart);
        },
        updateColumns: function (plottable) {
          console.log("SettingsView.updateColumns", plottable);

          if (!_.isEmpty(plottable) && !_.isEmpty(plottable.columns)) {
            var $input = this.$("select[name=xaxis]");
            $input.html("<option value=''>Incremental</option>");
            _.each(plottable.columns, function (item) {
              var $opt = $("<option>").val(item.column).text(item.column);
              if (this.xaxis && this.xaxis === item.column) {
                $opt.attr("selected", "selected");
              }
              $input.append($opt)
            }, this);
          }
        },
        addChartListItem: function (model) {
          this.$chartList.find(".empty").remove();
          this.subviews[model.cid] = new ChartListItem({model: model});
          this.$chartList.append(this.subviews[model.cid].render().el);
        },
        render: function () {
          this.$el.html(this.template({chartModes: this.chartModes, chart: this.model}));
          this.$(".chart-modes .btn:first").addClass("btn-primary");
          this.showLegend = this.$(".chart-show-legend").is(":checked");

          // need to reference chart list dropdown here because bootstrap dropdown fix
          // deattached the element from current dom
          this.$chartList = this.$(".dropdown-menu");

          return this;
        },
        loadChartMenu: function () {
          console.log("loadChartMenu", this.charts);
          var $ddmenu = this.$(".dropdown-menu");
          $ddmenu.html("");
          var that = this;
          this.charts.fetch({
            reset: true, success: function (models, data) {
              models.each(that.addChartListItem);
              if (data.length === 0) {
                $ddmenu.html($("<li class=empty><a>Empty</a></li>"));
              }
            }
          });
        },
        loadChart: function (chart) {
          console.log("settings.loadChart", chart.toJSON());

          // store chart instance
          this.model = chart;
          this.xaxis = null;

          // get chart settings
          var settings = chart.get("settings");

          // set chart mode setting
          if (settings && settings.chartMode) {
            var mode = settings.chartMode;
            var parse = this.parseModeTitle;
            this.$(".chart-modes .btn").removeClass("btn-primary").filter(function () {
              return parse($(this).data("mode")) === mode;
            }).addClass("btn-primary");
            this.chartMode = mode;
          }

          // set show legend setting
          this.$(".chart-show-legend").prop("checked", !!settings.showLegend);
          this.showLegend = !!settings.showLegend;

          // set chart title
          this.$(".chartTitle").val(chart.get("title"));

          // set sharing settings
          this.$(":input[name=public]").prop("checked", !!chart.get("public"));
          this.$(":input[name=team_shared]").prop("checked", !!chart.get("team_shared"));
          this.toggleEmbedURL(chart);

          // set xaxis setting
          if (!_.isEmpty(settings.xaxis)) {
            this.xaxis = settings.xaxis;
            this.$("select[name=xaxis] option").each(function () {
              if ($(this).val() === settings.xaxis) {
                $(this).attr("selected", "selected");
              }
            });
          }

          // trigger query events
          Events.trigger("query:set", chart.get("query"));
          Events.trigger("query:execute");
        },
        toggleEmbedURL: function (chart) {
          if (!!chart.get("public")) {
            this.$(".embed-url-wrapper").find(":input:first").val(chart.getEmbedURL());
            this.$(".embed-url-wrapper").show();
          } else {
            this.$(".embed-url-wrapper").hide();
          }
        }
      });
      return SettingsView;
    });
