/*global define*/
define([
  'jquery',
  'underscore',
  'backbone',
  'tpls',
  'baseview',
  'c3',
  'models/embed',
  'models/output/viz/chart',
  'moment'
], function ($, _, Backbone, Templates, BaseView, c3, Embed, Chart, Moment) {
  'use strict';

  var EmbedApp = BaseView.extend({
    el: '#embed-app',
    template: Templates["embed"],
    errorTemplate: _.template("<div class='alert'>Error: <%= msg %></div>"),
    initialize: function () {
      _.bindAll(this, "drawChart", "error");
    },
    render: function (id) {
      this.$el.html(this.template({}));

      var drawChart = this.drawChart;
      var error = this.error;
      var chart = new Chart({id: id});
      chart.url = "/embed/chart/" + id;
      chart.fetch({
        success: function (chartModel, chartData) {
          var embed = new Embed({id: id});
          embed.fetch({
            success: function (model, embedData) {
              drawChart(embedData, chartData, chartModel);
            },
            error: function () {
              error("Failed to load chart data.");
            }
          });
        },
        error: function () {
          error("Chart not found.");
        }
      });
    },
    error: function (msg) {
      this.$el.html(this.errorTemplate({msg: msg}));
    },
    getPlottableData: function (data) {

      // default values
      var plottableData = [],
          columns = [];

      if (data.length > 0) {

        // get plottable columns
        columns = this.getPlottableColumns(data);

        _.each(columns, function (item) {

          // get all values by column name
          var columnData = _.pluck(data, item.column);

          // check if column is date and needs to be parsed
          if (item.type === "date") {
            columnData = _.map(columnData, function (val) {
              return Moment(val);
            });
          }

          var escapedColumnName = item.column.replace(/([^a-zA-Z0-9_-])/g, '');
          columnData.unshift(escapedColumnName);
          plottableData.push(columnData);
        });
      }
      return {data: plottableData, columns: columns};
    },
    getPlottableColumns: function (data) {
      if (data.length > 0) {
        // check first row for plottable data
        var columns = _.map(data[0], function (v, k) {
          if (_.isNumber(v)) {
            return {type: "string", column: k};
          }

          var date = Moment(new Date(v));
          if (date.isValid()) {
            return {type: "date", column: k};
          }
        }, this);

        return _.compact(columns);
      }
      return [];
    },
    drawChart: function (embedData, chartData) {

      if (_.isEmpty(embedData) || _.isEmpty(chartData) || !_.isObject(chartData.settings)) {
        this.error("Chart data is incomplete.");
        return;
      }

      var plottable = this.getPlottableData(embedData.rows);

      // set data settings
      var dataSettings = {};
      dataSettings.columns = plottable.data;
      dataSettings.type = chartData.settings.chartMode;

      // get xaxis and update chart settings
      var axisSettings = {};
      var xAxis = chartData.settings.xaxis;
      if (!_.isEmpty(xAxis)) {

        // set xaxis settings
        dataSettings.x = xAxis;

        // find column in plottable columns
        var column = _.findWhere(plottable.columns, {column: xAxis});

        // check if xAxis could be timeseries
        if (!_.isEmpty(column) && column.type === "date") {
          axisSettings.x = {
            type: 'timeseries',
            tick: {
              format: '%Y-%m-%d',
              culling: {
                max: 5
              }
            }
          };
        }

      }

      // create chart
      this.chart = c3.generate({
        bindto: this.$(".chart").get(0),
        data: dataSettings,
        color: {
          pattern: ["#8EB746", "#3E2C52", "#E84B0C", "#D3D3D3"]
        },
        legend: {
          show: chartData.settings.showLegend
        },
        axis: axisSettings
      });
    }
  });

  return EmbedApp;
});